(function () {

/* Imports */
var Meteor = Package.meteor.Meteor;
var global = Package.meteor.global;
var meteorEnv = Package.meteor.meteorEnv;
var ECMAScript = Package.ecmascript.ECMAScript;
var Accounts = Package['accounts-base'].Accounts;
var SandstormDb = Package['sandstorm-db'].SandstormDb;
var SandstormPermissions = Package['sandstorm-permissions'].SandstormPermissions;
var meteorInstall = Package.modules.meteorInstall;
var Buffer = Package.modules.Buffer;
var process = Package.modules.process;
var Symbol = Package['ecmascript-runtime'].Symbol;
var Map = Package['ecmascript-runtime'].Map;
var Set = Package['ecmascript-runtime'].Set;
var meteorBabelHelpers = Package['babel-runtime'].meteorBabelHelpers;
var Promise = Package.promise.Promise;

/* Package-scope variables */
var inMeteor, promiseToFuture, waitPromise, SANDSTORM_ALTHOME, SandstormBackend;

var require = meteorInstall({"node_modules":{"meteor":{"sandstorm-backend":{"sandstorm-backend.js":function(require){

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                     //
// packages/sandstorm-backend/sandstorm-backend.js                                                                     //
//                                                                                                                     //
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                                       //
// Sandstorm - Personal Cloud Sandbox                                                                                  //
// Copyright (c) 2015 Sandstorm Development Group, Inc. and contributors                                               //
// All rights reserved.                                                                                                //
//                                                                                                                     //
// Licensed under the Apache License, Version 2.0 (the "License");                                                     //
// you may not use this file except in compliance with the License.                                                    //
// You may obtain a copy of the License at                                                                             //
//                                                                                                                     //
//   http://www.apache.org/licenses/LICENSE-2.0                                                                        //
//                                                                                                                     //
// Unless required by applicable law or agreed to in writing, software                                                 //
// distributed under the License is distributed on an "AS IS" BASIS,                                                   //
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.                                            //
// See the License for the specific language governing permissions and                                                 //
// limitations under the License.                                                                                      //
                                                                                                                       //
var Capnp = Npm.require("capnp");                                                                                      // 17
var Backend = Capnp.importSystem("sandstorm/backend.capnp").Backend;                                                   // 18
var Crypto = Npm.require("crypto");                                                                                    // 19
var Future = Npm.require("fibers/future");                                                                             // 20
                                                                                                                       //
var inMeteorInternal = Meteor.bindEnvironment(function (callback) {                                                    // 22
  callback();                                                                                                          // 23
});                                                                                                                    // 24
                                                                                                                       //
inMeteor = function inMeteor(callback) {                                                                               // 26
  // Calls the callback in a Meteor context.  Returns a Promise for its result.                                        //
  return new Promise(function (resolve, reject) {                                                                      // 28
    inMeteorInternal(function () {                                                                                     // 29
      try {                                                                                                            // 30
        resolve(callback());                                                                                           // 31
      } catch (err) {                                                                                                  // 32
        reject(err);                                                                                                   // 33
      }                                                                                                                // 34
    });                                                                                                                // 35
  });                                                                                                                  // 36
};                                                                                                                     // 37
                                                                                                                       //
promiseToFuture = function promiseToFuture(promise) {                                                                  // 39
  var result = new Future();                                                                                           // 40
  promise.then(result["return"].bind(result), result["throw"].bind(result));                                           // 41
  return result;                                                                                                       // 42
};                                                                                                                     // 43
                                                                                                                       //
waitPromise = function waitPromise(promise) {                                                                          // 45
  return promiseToFuture(promise).wait();                                                                              // 46
};                                                                                                                     // 47
                                                                                                                       //
SANDSTORM_ALTHOME = Meteor.settings && Meteor.settings.home;                                                           // 49
                                                                                                                       //
SandstormBackend = function SandstormBackend(db, backendCap) {                                                         // 51
  this._db = db;                                                                                                       // 52
  this._backendCap = backendCap;                                                                                       // 53
};                                                                                                                     // 54
                                                                                                                       //
SandstormBackend.prototype.cap = function () {                                                                         // 56
  return this._backendCap;                                                                                             // 57
};                                                                                                                     // 58
                                                                                                                       //
SandstormBackend.prototype.deleteUser = function (userId) {                                                            // 60
  return waitPromise(this._backendCap.deleteUser());                                                                   // 61
};                                                                                                                     // 62
                                                                                                                       //
SandstormBackend.prototype.shutdownGrain = function (grainId, ownerId, keepSessions) {                                 // 64
  if (!keepSessions) {                                                                                                 // 65
    Sessions.remove({ grainId: grainId });                                                                             // 66
  }                                                                                                                    // 67
                                                                                                                       //
  var grain = this._backendCap.getGrain(ownerId, grainId).supervisor;                                                  // 69
  return grain.shutdown().then(function () {                                                                           // 70
    grain.close();                                                                                                     // 71
    throw new Error("expected shutdown() to throw disconnected");                                                      // 72
  }, function (err) {                                                                                                  // 73
                                                                                                                       //
    grain.close();                                                                                                     // 75
    if (err.kjType !== "disconnected") {                                                                               // 76
      throw err;                                                                                                       // 77
    }                                                                                                                  // 78
  });                                                                                                                  // 79
};                                                                                                                     // 80
                                                                                                                       //
SandstormBackend.prototype.deleteGrain = function (grainId, ownerId) {                                                 // 82
  // We leave it up to the caller if they want to actually wait, but some don't so we report                           //
  // exceptions.                                                                                                       //
  return this._backendCap.deleteGrain(ownerId, grainId)["catch"](function (err) {                                      // 85
    console.error("problem deleting grain " + grainId + ":", err.message);                                             // 86
    throw err;                                                                                                         // 87
  });                                                                                                                  // 88
};                                                                                                                     // 89
                                                                                                                       //
SandstormBackend.shouldRestartGrain = function (error, retryCount) {                                                   // 91
  // Given an error thrown by an RPC call to a grain, return whether or not it makes sense to try                      //
  // to restart the grain and retry. `retryCount` is the number of times that the request has                          //
  // already gone through this cycle (should be zero for the first call).                                              //
                                                                                                                       //
  return error.kjType === "disconnected" && retryCount < 1;                                                            // 96
};                                                                                                                     // 97
                                                                                                                       //
SandstormBackend.prototype.maybeRetryUseGrain = function (grainId, cb, retryCount, err) {                              // 99
  var _this = this;                                                                                                    // 100
  if (SandstormBackend.shouldRestartGrain(err, retryCount)) {                                                          // 101
    return inMeteor(function () {                                                                                      // 102
      return cb(_this.continueGrain(grainId).supervisor)["catch"](_this.maybeRetryUseGrain.bind(_this, grainId, cb, retryCount + 1));
    });                                                                                                                // 105
  } else {                                                                                                             // 106
    throw err;                                                                                                         // 107
  }                                                                                                                    // 108
};                                                                                                                     // 109
                                                                                                                       //
SandstormBackend.prototype.useGrain = function (grainId, cb) {                                                         // 111
  var _this2 = this;                                                                                                   // 111
                                                                                                                       //
  // This will open a grain for you, handling restarts if needed, and call the passed function with                    //
  // the supervisor capability as the only parameter. The callback must return a promise that used                     //
  // the supervisor, so that we can check if a disconnect error occurred, and retry if possible.                       //
  // This function returns the same promise that your callback returns.                                                //
  //                                                                                                                   //
  // This function is NOT expected to be run in a meteor context.                                                      //
                                                                                                                       //
  return inMeteor(function () {                                                                                        // 119
    return cb(_this2.continueGrain(grainId).supervisor)["catch"](_this2.maybeRetryUseGrain.bind(_this2, grainId, cb, 0));
  });                                                                                                                  // 122
};                                                                                                                     // 123
                                                                                                                       //
SandstormBackend.prototype.continueGrain = function (grainId) {                                                        // 125
  var grain = Grains.findOne(grainId);                                                                                 // 126
  if (!grain) {                                                                                                        // 127
    throw new Meteor.Error(404, "Grain Not Found", "Grain ID: " + grainId);                                            // 128
  }                                                                                                                    // 129
                                                                                                                       //
  // If a DevPackage with the same app ID is currently active, we let it override the installed                        //
  // package, so that the grain runs using the dev app.                                                                //
  var devPackage = DevPackages.findOne({ appId: grain.appId });                                                        // 133
  var isDev = void 0;                                                                                                  // 134
  var pkg = void 0;                                                                                                    // 135
  if (devPackage) {                                                                                                    // 136
    isDev = true;                                                                                                      // 137
    pkg = devPackage;                                                                                                  // 138
  } else {                                                                                                             // 139
    pkg = Packages.findOne(grain.packageId);                                                                           // 140
    if (!pkg) {                                                                                                        // 141
      throw new Meteor.Error(500, "Grain's package not installed", "Package ID: " + grain.packageId);                  // 142
    }                                                                                                                  // 144
  }                                                                                                                    // 145
                                                                                                                       //
  var manifest = pkg.manifest;                                                                                         // 147
  var packageId = pkg._id;                                                                                             // 148
                                                                                                                       //
  if (!("continueCommand" in manifest)) {                                                                              // 150
    throw new Meteor.Error(500, "Package manifest defines no continueCommand.", "Package ID: " + packageId);           // 151
  }                                                                                                                    // 153
                                                                                                                       //
  var result = this.startGrainInternal(packageId, grainId, grain.userId, manifest.continueCommand, false, isDev);      // 155
  result.packageSalt = isDev ? pkg._id : grain.packageSalt;                                                            // 157
  return result;                                                                                                       // 158
};                                                                                                                     // 159
                                                                                                                       //
SandstormBackend.prototype.startGrainInternal = function (packageId, grainId, ownerId, command, isNew, isDev) {        // 161
  // Starts the grain supervisor.  Must be executed in a Meteor context.  Blocks until grain is                        //
  // started. Returns a promise for an object containing two fields: `owner` (the ID of the owning                     //
  // user) and `supervisor` (the supervisor capability).                                                               //
                                                                                                                       //
  if (isUserExcessivelyOverQuota(Meteor.users.findOne(ownerId))) {                                                     // 166
    throw new Meteor.Error(402, "Cannot start grain because owner's storage is exhausted.\n" + "Please ask them to upgrade.");
  }                                                                                                                    // 170
                                                                                                                       //
  // Ugly: Stay backwards-compatible with old manifests that had "executablePath" and "args" rather                    //
  //   than just "argv".                                                                                               //
  if ("args" in command) {                                                                                             // 174
    if (!("argv" in command)) {                                                                                        // 175
      command.argv = command.args;                                                                                     // 176
    }                                                                                                                  // 177
                                                                                                                       //
    delete command.args;                                                                                               // 179
  }                                                                                                                    // 180
                                                                                                                       //
  if ("executablePath" in command) {                                                                                   // 182
    if (!("deprecatedExecutablePath" in command)) {                                                                    // 183
      command.deprecatedExecutablePath = command.executablePath;                                                       // 184
    }                                                                                                                  // 185
                                                                                                                       //
    delete command.executablePath;                                                                                     // 187
  }                                                                                                                    // 188
                                                                                                                       //
  return this._backendCap.startGrain(ownerId, grainId, packageId, command, isNew, isDev);                              // 190
};                                                                                                                     // 191
                                                                                                                       //
var storageUsageUnimplemented = false;                                                                                 // 193
                                                                                                                       //
SandstormBackend.prototype.updateLastActive = function (grainId, userId, identityId) {                                 // 195
  // Update the lastActive date on the grain, any relevant API tokens, and the user,                                   //
  // and also update the user's storage usage.                                                                         //
                                                                                                                       //
  var storagePromise = undefined;                                                                                      // 199
  if (this._db.isQuotaEnabled() && !storageUsageUnimplemented) {                                                       // 200
    storagePromise = globalBackend._backendCap.getUserStorageUsage(userId);                                            // 201
  }                                                                                                                    // 202
                                                                                                                       //
  var now = new Date();                                                                                                // 204
  if (Grains.update(grainId, { $set: { lastUsed: now } }) === 0) {                                                     // 205
    // Grain must have been deleted. Ignore.                                                                           //
    return;                                                                                                            // 207
  }                                                                                                                    // 208
                                                                                                                       //
  if (userId) {                                                                                                        // 210
    Meteor.users.update(userId, { $set: { lastActive: now } });                                                        // 211
  }                                                                                                                    // 212
                                                                                                                       //
  if (identityId) {                                                                                                    // 214
    Meteor.users.update({ _id: identityId }, { $set: { lastActive: now } });                                           // 215
    // Update any API tokens that match this user/grain pairing as well                                                //
    ApiTokens.update({ grainId: grainId, "owner.user.identityId": identityId }, { $set: { lastUsed: now } }, { multi: true });
  }                                                                                                                    // 220
                                                                                                                       //
  if (storagePromise) {                                                                                                // 222
    try {                                                                                                              // 223
      var ownerId = Grains.findOne(grainId).userId;                                                                    // 224
      var size = parseInt(waitPromise(storagePromise).size);                                                           // 225
      Meteor.users.update(ownerId, { $set: { storageUsage: size } });                                                  // 226
      // TODO(security): Consider actively killing grains if the user is excessively over quota?                       //
      //   Otherwise a constantly-active grain could consume arbitrary space without being stopped.                    //
    } catch (err) {                                                                                                    // 229
      if (err.kjType === "unimplemented") {                                                                            // 230
        storageUsageUnimplemented = true;                                                                              // 231
      } else {                                                                                                         // 232
        console.error("error getting user storage usage:", err.stack);                                                 // 233
      }                                                                                                                // 234
    }                                                                                                                  // 235
  }                                                                                                                    // 236
};                                                                                                                     // 237
                                                                                                                       //
function generateSessionId(grainId, userId, packageSalt, clientSalt) {                                                 // 239
  var sessionParts = [grainId, clientSalt];                                                                            // 240
  if (userId) {                                                                                                        // 241
    sessionParts.push(userId);                                                                                         // 242
  }                                                                                                                    // 243
                                                                                                                       //
  if (packageSalt) {                                                                                                   // 245
    sessionParts.push(packageSalt);                                                                                    // 246
  }                                                                                                                    // 247
                                                                                                                       //
  var sessionInput = sessionParts.join(":");                                                                           // 249
  return Crypto.createHash("sha256").update(sessionInput).digest("hex");                                               // 250
}                                                                                                                      // 251
                                                                                                                       //
SandstormBackend.prototype.openSessionInternal = function (grainId, userId, identityId, title, apiToken, cachedSalt) {
                                                                                                                       //
  // Start the grain if it is not running. This is an optimization: if we didn't start it here,                        //
  // it would start on the first request to the session host, but we'd like to get started before                      //
  // the round trip.                                                                                                   //
                                                                                                                       //
  var _continueGrain = this.continueGrain(grainId);                                                                    // 253
                                                                                                                       //
  var supervisor = _continueGrain.supervisor;                                                                          // 253
  var packageSalt = _continueGrain.packageSalt;                                                                        // 253
                                                                                                                       //
                                                                                                                       //
  this.updateLastActive(grainId, userId, identityId);                                                                  // 260
                                                                                                                       //
  cachedSalt = cachedSalt || Random.id(22);                                                                            // 262
  var sessionId = generateSessionId(grainId, userId, packageSalt, cachedSalt);                                         // 263
  var session = Sessions.findOne({ _id: sessionId });                                                                  // 264
  if (session) {                                                                                                       // 265
    // TODO(someday): also do some more checks for anonymous sessions (sessions without a userId).                     //
    if (session.identityId && session.identityId !== identityId || session.grainId !== grainId) {                      // 267
      var e = new Meteor.Error(500, "Duplicate SessionId");                                                            // 269
      console.error(e);                                                                                                // 270
      throw e;                                                                                                         // 271
    } else {                                                                                                           // 272
      return {                                                                                                         // 273
        supervisor: supervisor,                                                                                        // 274
        methodResult: {                                                                                                // 275
          sessionId: session._id,                                                                                      // 276
          title: title,                                                                                                // 277
          grainId: grainId,                                                                                            // 278
          hostId: session.hostId,                                                                                      // 279
          tabId: session.tabId,                                                                                        // 280
          salt: cachedSalt                                                                                             // 281
        }                                                                                                              // 275
      };                                                                                                               // 273
    }                                                                                                                  // 284
  }                                                                                                                    // 285
                                                                                                                       //
  // TODO(someday): Allow caller to specify the parent session from which to inherit the tab ID, or                    //
  //   something.                                                                                                      //
                                                                                                                       //
  session = {                                                                                                          // 290
    _id: sessionId,                                                                                                    // 291
    grainId: grainId,                                                                                                  // 292
    hostId: Crypto.createHash("sha256").update(sessionId).digest("hex").slice(0, 32),                                  // 293
    tabId: Crypto.createHash("sha256").update("tab:").update(sessionId).digest("hex").slice(0, 32),                    // 294
    timestamp: new Date().getTime(),                                                                                   // 295
    hasLoaded: false                                                                                                   // 296
  };                                                                                                                   // 290
                                                                                                                       //
  if (userId) {                                                                                                        // 299
    session.identityId = identityId;                                                                                   // 300
    session.userId = userId;                                                                                           // 301
  } else if (apiToken) {                                                                                               // 302
    session.hashedToken = apiToken._id;                                                                                // 303
  } else {}                                                                                                            // 304
  // Must be old-style sharing, i.e. !grain.private.                                                                   //
  // jscs:ignore disallowEmptyBlocks                                                                                   //
                                                                                                                       //
  Sessions.insert(session);                                                                                            // 308
                                                                                                                       //
  return {                                                                                                             // 310
    supervisor: supervisor,                                                                                            // 311
    methodResult: {                                                                                                    // 312
      sessionId: session._id,                                                                                          // 313
      title: title,                                                                                                    // 314
      grainId: grainId,                                                                                                // 315
      hostId: session.hostId,                                                                                          // 316
      tabId: session.tabId,                                                                                            // 317
      salt: cachedSalt                                                                                                 // 318
    }                                                                                                                  // 312
  };                                                                                                                   // 310
};                                                                                                                     // 321
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

}}}}},{"extensions":[".js",".json"]});
require("./node_modules/meteor/sandstorm-backend/sandstorm-backend.js");

/* Exports */
if (typeof Package === 'undefined') Package = {};
(function (pkg, symbols) {
  for (var s in symbols)
    (s in pkg) || (pkg[s] = symbols[s]);
})(Package['sandstorm-backend'] = {}, {
  SandstormBackend: SandstormBackend
});

})();

//# sourceMappingURL=sandstorm-backend.js.map
